/*
 * Modified MineSweeper game (original made by Scott Mitchell
 * http://www.eecs.qmul.ac.uk/~mmh/ItP/resources/MineSweeper/Notes.html)
 */
package hu.mit.bme.ftsrg.gameEnvironment.mineSweeper;

import java.io.*;

import hu.mit.bme.ftsrg.gameEnvironment.SocketCommunication;

/**
 * A basic implementation of the MineSweeper game, using a TextBoard for display purposes and a StringTokenizer to parse
 * commands entered at the keyboard. No support yet for anything fancy like high score tables (or graphics :) We do,
 * however, time how long it takes you to finish the game.
 */
public class MineSweeper {

	// The game board
	private Board board;

	// Tokenizer for commands
	private StreamTokenizer tok;

	// Various flags used by play() and doCommand()
	private boolean done, quit, win;

	// Contents of last cell revealed
	private int lastCell;

	private SocketCommunication socketComm;

	/**
	 * Create and initialise a new game of the given dimensions.
	 */
	public MineSweeper(int width, int height, int mines,
			SocketCommunication socketComm) {

		this.socketComm = socketComm;
		// Create the game board
		board = new TextBoard(width, height, mines);

		// Set up the command tokenizer
		tok = new StreamTokenizer(socketComm.in);

		done = win = quit = false;
	}

	/**
	 * Main game loop.
	 */
	public void play() throws IOException {
		// Start the clock
		long startTime = System.currentTimeMillis();

		// Loop until game over, for whatever reason
		while (!done) {

			// Redraw board
			board.draw(socketComm);

			// Get next command
			socketComm.write("Command: ");
			tok.nextToken();

			// Dispatch on command
			switch (tok.ttype) {
			case StreamTokenizer.TT_WORD:
				doCommand();
				break;
			case StreamTokenizer.TT_EOL:
				continue;
			case StreamTokenizer.TT_EOF:
				done = quit = true;
				break;
			default:
				socketComm.writeLine("Unknown command -- try 'help'");
			}

			// Check whether the game is up. If the number of unknown cells = the number of mines, you must have won (by
			// finding all the mines). If, on the other hand, you revealed a mine, time to die :(
			if (board.getUnknown() == board.getMines()) {
				done = win = true;
			} else if (lastCell == Board.MINE) {
				done = true;
			}

			// Remove remaining garbage from the input stream.
			socketComm.readLine();
		}

		// Game over, one way or another. Figure out how long it took
		long elapsedTime = System.currentTimeMillis() - startTime;

		// Reveal everything, just so the stats are correct
		for (int i = 0; i < board.getWidth(); i++) {
			for (int j = 0; j < board.getHeight(); j++) {
				board.reveal(i, j);
			}
		}

		// Redraw the board and print out some messages
		board.draw(socketComm);
		if (win) {
			socketComm.writeLine("Congratulations -- you found all the mines!");
		} else if (!quit) {
			socketComm.writeLine("Bad luck -- you stepped on a mine!");
		}
		socketComm.writeLine("Total game time: " + (elapsedTime / 1000)
				+ " seconds");
	}

	/**
	 * Process commands. 'Help' gives a summary of the huge range of available commands.
	 */
	private void doCommand() throws IOException {
		int x, y;

		if (tok.sval.equals("reveal")) {
			// Reveal a cell (and perhaps some around it)
			tok.nextToken();
			x = (int) tok.nval;
			tok.nextToken();
			y = (int) tok.nval;
			lastCell = board.reveal(x, y);
			if (lastCell == 0) {
				board.revealMore(x, y);
			}
		} else if (tok.sval.equals("mark")) {
			// Mark a cell
			tok.nextToken();
			x = (int) tok.nval;
			tok.nextToken();
			y = (int) tok.nval;
			board.mark(x, y);
		} else if (tok.sval.equals("unmark")) {
			// Unmark a cell
			tok.nextToken();
			x = (int) tok.nval;
			tok.nextToken();
			y = (int) tok.nval;
			board.unmark(x, y);
		} else if (tok.sval.equals("help")) {
			// Print some help
			socketComm.writeLine("reveal <x> <y>");
			socketComm.writeLine("mark <x> <y>");
			socketComm.writeLine("unmark <x> <y>");
			socketComm.writeLine("help");
			socketComm.writeLine("quit");
		} else if (tok.sval.equals("quit")) {
			// Quit and die
			quit = done = true;
		} else {
			socketComm.writeLine("Unknown command -- try 'help'");
		}
	}
}
